/******************************************************************************/
/*Files to Include                                                            */
/******************************************************************************/

#if defined(__XC)
    #include <xc.h>         /* XC8 General Include File */
#elif defined(HI_TECH_C)
    #include <htc.h>        /* HiTech General Include File */
#endif

#include <stdint.h>         /* For uint8_t definition */
#include <stdbool.h>
#include <string.h>        /* For true/false definition */

#include "system.h"
#include "user.h"

/******************************************************************************/
/* Interrupt Routines                                                         */
/******************************************************************************/
static volatile uint8_t __dummy_byte;
static volatile uint8_t __rx_data_count;
static volatile uint8_t __index;

static void inline __handle_ssp_isr(void);
static void inline __handle_bcl_isr(void);
static void inline __handle_adc_isr(void);

void interrupt isr(void)
{
    if (PIR1bits.SSP1IF) {
        __handle_ssp_isr();
    } else if(PIR2bits.BCL1IF) {
        __handle_bcl_isr();
    } else if(PIR1bits.ADIF) {
        __handle_adc_isr();
    } else {
        /* Unmasked interrupt vector */
        while (1) {
            /* Wait for WDT reset */
            asm("NOP");
        }
    }    
}
static void inline __handle_adc_isr(void)
{
    /* Store ADC result */
    adc.data = ADRES;
    
    ANSELA = 0;
    ANSELC = 0;
    
    ADCON0bits.ADON = 0;
    
    /* Clear interrupt flag */
    PIR1bits.ADIF = 0;
}

static void inline __handle_bcl_isr(void) {
    /* Clear buffer */
    __dummy_byte = SSPBUF;

    /* Clear bus collision interrupt flag */
    PIR2bits.BCL1IF = 0;
    
    /* Release clock line */
    SSP1CON1bits.CKP = 1;

}


/**
 * @brief __handle_ssp_isr
 * Interrupt routine for handle I2C communication
 */
static void inline __handle_ssp_isr(void)
{
    
    if (SSPSTATbits.R_nW) {
        /* We have READ */

        __dummy_byte = SSPBUF;

        switch(__index++) {
        case 0x01:
            SSPBUF = registers.tris;
            break;
            
        case 0x02:
            SSPBUF = registers.lat;
            break;
            
        case 0x03:
            SSPBUF = registers.port;
            break;
            
        case 0x04:
            SSPBUF = registers.pullup;
            break;
            
        case 0x10:
            SSPBUF = (adc.index == 0x10) ? adc.data_l : 0xFF;
            break;
            
        case 0x11:
            if(adc.index == 0x10)
                SSPBUF = adc.data_h;
            else
                SSPBUF = (adc.index == 0x11) ? adc.data_l : 0xFF;
            break;
            
        case 0x12:
            if(adc.index == 0x11)
                SSPBUF = adc.data_h;
            else
                SSPBUF = (adc.index == 0x12) ? adc.data_l : 0xFF;
            break;
            
        case 0x13:
            SSPBUF = (adc.index == 0x12) ? adc.data_h : 0xFF;
            break;
            
        case 0x16:
            SSPBUF = (adc.index == 0x16) ? adc.data_l : 0xFF;;
            break;
            
        case 0x17:
            if(adc.index == 0x16)
                SSPBUF = adc.data_h;
            else
                SSPBUF = (adc.index == 0x17) ? adc.data_l : 0xFF;
            break;
            
        case 0x18:
            SSPBUF = (adc.index == 0x17) ? adc.data_h : 0xFF;
            break;
            
        case 0x20:
            SSPBUF = registers.id;
            break;
            
        case 0x21:
            SSPBUF = registers.temperature[3];
            break;
            
        case 0x22:
            SSPBUF = registers.temperature[2];
            break;
            
        case 0x23:
            SSPBUF = registers.temperature[1];
            break;
            
        case 0x24:
            SSPBUF = registers.temperature[0];
            break;
            
        case 0xF0:
            SSPBUF = registers.address >> 1;
            break;
            
        default:
            SSPBUF = 0xFF;
            break;
        }

        /* Release clock line */
        SSPCON1bits.CKP = 1;

    } else {
        /* We have WRITE */
        if (!SSPSTATbits.D_nA) {
            /* Last was address*/
            
            __dummy_byte = SSPBUF;
            __rx_data_count = 0;
            
            /* Release clock */
            SSPCON1bits.CKP = 1;
        } else {
            /* Last was data */
            
            if(!__rx_data_count++) {
                /* This was register pointer(command) */
                __index = SSPBUF;
                
                /* Check for analog read commands */
                switch(__index) {
                    case 0x10:
                        TRISAbits.TRISA0 = 1;
                        ANSELAbits.ANSA0 = 1;
                        registers.tris |= 0x01;
                        ADCON0bits.CHS = 0x00;
                        adc.index = __index;
                        ADCON0bits.ADON = 1;
                        __delay_us(15);
                        ADCON0bits.GO = 1;
                        break;
                        
                    case 0x11:
                        TRISAbits.TRISA1 = 1;
                        ANSELAbits.ANSA1 = 1;
                        registers.tris |= 0x02;
                        ADCON0bits.CHS = 0x01;
                        adc.index = __index;
                        ADCON0bits.ADON = 1;
                        __delay_us(15);
                        ADCON0bits.GO = 1;
                        break;
                        
                    case 0x12:
                        TRISAbits.TRISA2 = 1;
                        ANSELAbits.ANSA2 = 1;
                        ADCON0bits.CHS = 0x02;
                        registers.tris |= 0x04;
                        adc.index = __index;
                        ADCON0bits.ADON = 1;
                        __delay_us(15);
                        ADCON0bits.GO = 1;
                        break;
                        
                    case 0x16:
                        TRISCbits.TRISC2 = 1;
                        ANSELCbits.ANSC2 = 1;
                        ADCON0bits.CHS = 0x06;
                        registers.tris |= 0x20;
                        adc.index = __index;
                        ADCON0bits.ADON = 1;
                        __delay_us(15);
                        ADCON0bits.GO = 1;
                        break;
                        
                    case 0x17:
                        TRISCbits.TRISC3 = 1;
                        ANSELCbits.ANSC3 = 1;
                        ADCON0bits.CHS = 0x07;
                        registers.tris |= 0x40;
                        adc.index = __index;
                        ADCON0bits.ADON = 1;
                        __delay_us(15);
                        ADCON0bits.GO = 1;
                        break;
                        
                    default:
                        break;
                }
                
            } else {                
                switch(__index++) {
                    case 0x01:
                        registers.tris = ((SSPBUF | 0x08 ) & 0x7F);
                        break;
                        
                    case 0x02:
                        registers.lat = SSPBUF & 0x77;
                        break;
                        
                    case 0x04:
                        registers.pullup = SSPBUF & 0x1F;
                        break;
                        
                    case 0xF0:
                        __dummy_byte = SSPBUF;
                        if(!PORTAbits.RA5 && __dummy_byte > 0x2 && __dummy_byte < 0x78)
                            registers.address = __dummy_byte << 1;
                        break;
                        
                    default:
                        __dummy_byte = SSPBUF;
                        break;
                }                
            }

            /* Check for write collision */
            if (SSP1CON1bits.WCOL) {
                SSP1CON1bits.WCOL = 0;
                __dummy_byte = SSPBUF;
            }

            /* Release clock line */
            SSPCON1bits.CKP = 1;
        }
    }

    /* Clear interrupt flag */
    PIR1bits.SSP1IF = 0;

}
