/******************************************************************************/
/* Files to Include                                                           */
/******************************************************************************/

#if defined(__XC)
    #include <xc.h>         /* XC8 General Include File */
#elif defined(HI_TECH_C)
    #include <htc.h>        /* HiTech General Include File */
#endif

#include <stdint.h>         /* For uint8_t definition */
#include <stdbool.h>        /* For true/false definition */
#include <string.h>

#include "user.h"
#include "memory.h"

/******************************************************************************/
/* User Functions                                                             */
/******************************************************************************/

static void inline __initGPIO(void);
static void inline __initMSSP(void);
static void inline __initI2C(void);
static void inline __initADC(void);

static void inline __setTris(void);
static void inline __getTris(void);
static void inline __setLat(void);
static void inline __getLat(void);
static void inline __getPort(void);
static void inline __setPullups(void);
static void inline __getPullups(void);
static void inline __updateAddress(void);

static uint8_t __spiReadByte(void);

static uint8_t __getI2CAddress(void);
static void __setI2CAddress(uint8_t address);

void spiReadWord(uint8_t *data);
void update(void);




/**
 * Initialize user programs
 */
void InitApp(void)
{
    /* Initialize */
    __initGPIO();
    __initI2C();
    __initADC();
    
    registers.id = DEVICE_ID;

    /* Enable interrupts */
    PIR1bits.SSP1IF = 0;
    PIE1bits.SSP1IE = 1;
    
    PIR1bits.ADIF = 0;
    PIE1bits.ADIE = 1;
    
    INTCONbits.PEIE = 1;
    INTCONbits.GIE = 1;
}

/**
 * Read 4-bytes (1 word) from MAX61855
 * @param data
 */
void spiReadWord(uint8_t *data)
{
    LATAbits.LATA4 = 0;
    *(data + 3) = __spiReadByte();
    *(data + 2) = __spiReadByte();
    *(data + 1) = __spiReadByte();
    *data = __spiReadByte();
    LATAbits.LATA4 = 1;
}

/**
 * Update registers
 */
void update(void)
{
    uint8_t temp[4];
    
    __setTris();
    __setLat();
    __getPort();
    __setPullups();
    
    spiReadWord(temp);
    memcpy(registers.temperature, temp, 4);
    
    __updateAddress();
}

/**
 * Store new address, if register is different
 */
static void inline __updateAddress(void)
{
    if(SSPADD != registers.address) {
        __setI2CAddress(registers.address);
        SSPADD = registers.address;
    }
    
}

/**
 * Read single byte from SPI slave device
 * @return Read byte
 */
static uint8_t __spiReadByte(void)
{
    uint8_t data = 0;
    uint8_t i = 7;
    
    do {
        LATCbits.LATC5 = 1;   // 2 instructions
        data |= (PORTCbits.RC4 << i);   // 11 instructions
        LATCbits.LATC5 = 0;   // 2 instructions
        asm("NOP");
        asm("NOP");        
    }while(i--);
    
    return data;
}

/**
 * Initialize ADC
 */
static void inline __initADC(void)
{
    ADCON1bits.ADFM = 1;    // Right justified
    ADCON1bits.ADCS = 5;    // Fosc/16 -> 1us
}

/**
 * Initialize MSSP as I2C slave device
 */
static void inline __initI2C(void)
{
    SSP1STAT = 0x80; // I2C running at 100kHz
    SSP1CON1 = 0x36; // i2c slave mode, 7-bit addressing without START and STOP bit interrupts

    SSP1CON2bits.SEN = 1; // Enable clock stretching 

    SSP1CON3bits.BOEN = 1; // Enable buffer overwrite
    SSP1CON3bits.SDAHT = 1; // Enable 300 ns SDA hold time
    SSP1CON3bits.SBCDE = 1; // Enable bus collision interrupt
    
    SSP1CON3bits.PCIE = 1; // Enable Stop condition interrupt
    
    if(!PORTAbits.RA5) {
        registers.address = ADDR_DFL << 1;
    } else {
        registers.address = __getI2CAddress();
        if ((registers.address >> 1) < 0x03 || (registers.address >> 1) > 0x77) {
            registers.address = ADDR_DFL << 1;
            __setI2CAddress(registers.address);
        }
    }
    SSPADD = registers.address;
}

/**
 * Initialize IO
 */
static void inline __initGPIO(void)
{
    /* Disable all ADC functions */
    ANSELA = 0;
    ANSELC = 0;
    
    
    
    /* Setup SPI interface */
    TRISAbits.TRISA4 = 0;       // SS
    TRISCbits.TRISC4 = 1;       // MISO
    TRISCbits.TRISC5 = 0;       // SCK
    LATAbits.LATA4 = 1;
    
    /* Enable weak pull-ups */
    OPTION_REGbits.nWPUEN = 0;
    WPUA = 0;
    
    /* Setup PROG */
    TRISAbits.TRISA5 = 1;
    WPUAbits.WPUA5 = 1;
    
    __getTris();
    __getLat();
    __getPullups();    
    
}


static void inline __setTris(void)
{
    TRISAbits.TRISA0 = (registers.tris & 0x01) ? 1 : 0;
    TRISAbits.TRISA1 = (registers.tris & 0x02) ? 1 : 0;
    TRISAbits.TRISA2 = (registers.tris & 0x04) ? 1 : 0;
    TRISAbits.TRISA5 = (registers.tris & 0x10) ? 1 : 0;
    TRISCbits.TRISC2 = (registers.tris & 0x20) ? 1 : 0;
    TRISCbits.TRISC3 = (registers.tris & 0x40) ? 1 : 0;
}

static void inline __getTris(void)
{
    registers.tris = 0x00;
    
    registers.tris |= (TRISAbits.TRISA0 << 0);
    registers.tris |= (TRISAbits.TRISA1 << 1);
    registers.tris |= (TRISAbits.TRISA2 << 2);
    registers.tris |= (1 << 3);
    registers.tris |= (TRISAbits.TRISA5 << 4);
    registers.tris |= (TRISCbits.TRISC2 << 5);
    registers.tris |= (TRISCbits.TRISC3 << 6);
}
static void inline __setLat(void)
{
    LATAbits.LATA0 = (registers.lat & 0x01) ? 1 : 0;
    LATAbits.LATA1 = (registers.lat & 0x02) ? 1 : 0;
    LATAbits.LATA2 = (registers.lat & 0x04) ? 1 : 0;
    LATAbits.LATA5 = (registers.lat & 0x10) ? 1 : 0;
    LATCbits.LATC2 = (registers.lat & 0x20) ? 1 : 0;
    LATCbits.LATC3 = (registers.lat & 0x40) ? 1 : 0;
}

static void inline __getLat(void)
{
    registers.lat = 0x00;
    
    registers.lat |= (LATAbits.LATA0 << 0);
    registers.lat |= (LATAbits.LATA1 << 1);
    registers.lat |= (LATAbits.LATA2 << 2);
    registers.lat |= (0 << 3);
    registers.lat |= (LATAbits.LATA5 << 4);
    registers.lat |= (LATCbits.LATC2 << 5);
    registers.lat |= (LATCbits.LATC3 << 6);
}
static void inline __getPort(void)
{
    uint8_t port = 0x00;
    
    port |= (PORTAbits.RA0 << 0);
    port |= (PORTAbits.RA1 << 1);
    port |= (PORTAbits.RA2 << 2);
    port |= (PORTAbits.RA3 << 3);
    port |= (PORTAbits.RA5 << 4);
    port |= (PORTCbits.RC2 << 5);
    port |= (PORTCbits.RC3 << 6);    
    
    registers.port = port;
}

static void inline __setPullups(void)
{
    WPUAbits.WPUA0 = (registers.pullup & 0x01) ? 1 : 0;
    WPUAbits.WPUA1 = (registers.pullup & 0x02) ? 1 : 0;
    WPUAbits.WPUA2 = (registers.pullup & 0x04) ? 1 : 0;
    WPUAbits.WPUA3 = (registers.pullup & 0x08) ? 1 : 0;
    WPUAbits.WPUA5 = (registers.pullup & 0x10) ? 1 : 0;
}

static void inline __getPullups(void)
{
    registers.pullup = 0x00;
    
    registers.pullup |= (WPUAbits.WPUA0 << 0);
    registers.pullup |= (WPUAbits.WPUA1 << 1);
    registers.pullup |= (WPUAbits.WPUA2 << 2);
    registers.pullup |= (WPUAbits.WPUA3 << 3);
    registers.pullup |= (WPUAbits.WPUA5 << 4);
}

static uint8_t __getI2CAddress(void)
{
    return FLASH_ReadWord(ADDR_MEM) & 0xFF;
}

static void __setI2CAddress(uint8_t address)
{
    uint16_t temp[ERASE_FLASH_BLOCKSIZE];
    
    FLASH_WriteWord(ADDR_MEM, temp, address);
}